#include <math.h>
#include <fcntl.h>
#include <linux/fb.h>
#include <pthread.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdarg.h>
#include <stropts.h>
#include <time.h>
#include <semaphore.h>

#include "snes9x.h"
#include "memmap.h"
#include "apu.h"
#include "gfx.h"
#include "soundux.h"
#include "snapshot.h"
#include "gp2x_cpuctrl.h"
#include "gp2x_asmutils.h"

extern "C"
{
#include "gp2x_minimal.h"
}

#include "menu/menusys.h"
#include "menu/image.h"

#define EMUVERSION "SquidgeSNES V0.392"

#define NUMSAMPLES (256)

static unsigned int pause_to_menu = 0;
static char cmdline_rom[256];
static int oldscaley = 0;
static hm::Image *gamebg = NULL;
static int audiofd = -1;
static pthread_t soundthread_p = (pthread_t) -1;
static struct menureturn local_mr;
static struct timeval noticeMsgTime = { 0, 0 };	// when started showing
static char displaytext[256];
static short samplebuffer[2*44100/50];
static int   samplebuffer_bytes = 0;
static sem_t samplebuffer_sem;

extern char **g_argv;

int yo_rand_val=0;
int32 gp32_fastmode = 1;
int gp32_8bitmode = 0;
int32 gp32_ShowSub = 0;
int gp32_fastsprite = 0;
uint8 *vrambuffer;

int sndvolL, sndvolR;
int squidgetranshack = 0;

static int frames_done = 0, frames_shown = 0, target_frametime, thissec;
static int shownfps = 0, shownfps_e = 0;


//---------------------------------------------------------------------------

void checkbatt (void);

int yo_rand(void)
{
     yo_rand_val++;
     return yo_rand_val;
}

void S9xExit ()
{
}

#if 0
void drawRect (unsigned short *vram, int x, int y, int w, int h)
{
	for (int y1 = y; y1 < y+h; y1 ++)
		for (int x1 = x; x1 < x+w; x1 ++)
			vram[(y1*320)+x1] ^= 0xFFFF;
}

void drawBatteryIcon (void)
{
	unsigned short *vram = (unsigned short *) gp2x_video_RGB[0].screen16;
	int battX = 148, battY = 212, battWidth = 26, battHeight = 10;

	drawRect(vram, battX, battY, battWidth-2, battHeight);
	drawRect(vram, battX+1, battY+1, battWidth-4, battHeight-2);
	// nib
	drawRect(vram, (battX+(battWidth-2)), battY+2, 2, battHeight-4);
}
#endif


extern "C" void S9xSetPalette ()
{
	if (gp32_8bitmode)
	{	
		unsigned int R=0,G=0,B=0;
		for (int i = 0; i < 256; i++)
		{	
			R = PPU.CGDATA [i] & 0x1f;
			G = (PPU.CGDATA [i] >> 5 ) & 0x1f;
			B = (PPU.CGDATA [i] >> 10) & 0x1f;
			R=(R|(R<<5))>>2; // 0x000R RRRR
			// 0xRRRR RRRR RR
			G=(G|(G<<5))>>2;
			B=(B|(B<<5))>>2;
			gp2x_video_RGB_color8(R,G,B,i);		
		}
		gp2x_video_RGB_setpalette();
	}
}

extern "C" void S9xExtraUsage ()
{
}
	
extern "C" void S9xParseArg (char **argv, int &index, int argc)
{
}

extern "C" bool8 S9xOpenSnapshotFile (const char *fname, bool8 read_only, STREAM *file)
{
	if (read_only)
	{
		if (*file = OPEN_STREAM(fname,"rb")) 
			return(TRUE);
	}
	else
	{
		if (*file = OPEN_STREAM(fname,"w+b")) 
			return(TRUE);
	}

	return (FALSE);	
}
	
extern "C" void S9xCloseSnapshotFile (STREAM file)
{
	CLOSE_STREAM(file);
	sync();
}

extern "C" void S9xMessage (int /* type */, int /* number */, const char *message)
{
	printf ("%s\n", message);
}

void erk (void)
{
	S9xMessage (0,0, "Erk!");
}

extern "C" void S9xLoadSDD1Data (void)
{
	S9xMessage (0,0,"load sdd1data");
	Memory.FreeSDD1Data();
}

static void simpleWait(int thissec, int lim_time)
{
	struct timeval tval;

	spend_cycles(1024);
	gettimeofday(&tval, 0);
	if(thissec != tval.tv_sec) tval.tv_usec+=1000000;

	while(tval.tv_usec < lim_time)
	{
		spend_cycles(1024);
		gettimeofday(&tval, 0);
		if(thissec != tval.tv_sec) tval.tv_usec+=1000000;
	}
}

extern "C" bool8 S9xInitUpdate()
{
	static int needfskip = 0;
	struct timeval tval;

	GFX.Screen = (uint8 *) gp2x_video_RGB[0].screen16;

	if (!local_mr.wantscale)
	{
		GFX.Screen += gp32_8bitmode ? 32 : 64;

		if (PPU.ScreenHeight != SNES_HEIGHT_EXTENDED)
		{
			GFX.Screen += gp32_8bitmode ? 320*8 : 640*8;
		}
	}

	gettimeofday(&tval, 0);
	if (tval.tv_sec != thissec)
	{
		thissec = tval.tv_sec;
		shownfps = frames_shown;
		shownfps_e = frames_done;
		frames_shown = 0;
		/*int delta = Memory.ROMFramesPerSecond - frames_done;*/
		if (local_mr.fskip < 0 && Memory.ROMFramesPerSecond - frames_done == 1)
			 frames_done = -1;
		else frames_done = 0;
	}
	frames_done++;


	if (local_mr.fskip == 0)
		return TRUE; // No fskip wanted
	else if (local_mr.fskip < 0)
	{
		// auto frameskip
		int lim_time = (frames_done) * target_frametime;
		if(tval.tv_usec > lim_time)
			return FALSE;
	}
	else if (needfskip == 0)
	{
		needfskip = local_mr.fskip;
		return TRUE;
	}
	else // if (needfskip)
	{
		needfskip--;
		return FALSE;
	}

	return TRUE;
}


extern "C" bool8 S9xDeinitUpdate (int Width, int Height, bool8)
{
	struct timeval tval;

	frames_shown++;
	gettimeofday(&tval, 0);

	if (local_mr.wantscale)
	{
		// Height can change, so change the scaling factor depending on it. Width should never change... 
		if (oldscaley != Height)
		{
			oldscaley = Height;
			gp2x_video_RGB_setscaling (Width, Height);
		}
	}

	if (local_mr.wantfps)
	{
		gp2x_printf(NULL, 0, 0, " %i/%i  ", shownfps, shownfps_e);
	}

	if(noticeMsgTime.tv_sec) {
		if((tval.tv_sec*1000000+tval.tv_usec) - (noticeMsgTime.tv_sec*1000000+noticeMsgTime.tv_usec) > 2000000) { // > 2.0 sec
			noticeMsgTime.tv_sec = noticeMsgTime.tv_usec = 0;
		} else {
			gp2x_printf(NULL, 160 - (strlen(displaytext)*3), 8, displaytext);      
		}
	}

	if((local_mr.apuenable && (!local_mr.apumute)) || local_mr.fskip < 0)
	{
		// usleep sleeps for ~20ms minimum, so it is not a solution here
		int lim_time = (frames_done) * target_frametime;
		if(thissec != tval.tv_sec) tval.tv_usec+=1000000;
		if(tval.tv_usec < lim_time)
		{
			// we are too fast
			simpleWait(thissec, lim_time);
		}
	}

	// check if battery icon needs to be on

	//gp2x_printf(NULL, 0,0,"LBV:%d BV:%d", local_mr.lowbattval, battvolts);

	/*if (local_mr.lowbattval)
	{
		checkbatt();

		if (battvolts <= local_mr.lowbattval)
		{			
			static int battontime = 0;

			battontime ++;
			if (battontime > 50)
			{
				drawBatteryIcon();
				if (battontime > 100)
					battontime = 0;		
			}
		}
	}*/

	if (local_mr.vsyncenable)
		gp2x_video_waitvsync();

	gp2x_video_RGB_flip(0);
}


extern "C" const char *S9xGetFilename (const char *ex) // ex is like ".srm"
{
	static char filename [PATH_MAX];
	char *p;

	strcpy (filename, Memory.ROMFilename);
	for(p = filename+strlen(filename)-1; p > filename+3; p--)
		if(*p == '\\' || *p == '/') break;
	memmove(filename, p+1, strlen(p)+1);
	for(p = filename+strlen(filename)-1; p > filename+3; p--)
		if(*p == '.' || *p == '\\' || *p == '/') break;
	strcpy(p, ex);

	return filename;
}


extern "C" uint32 S9xReadJoypad (int which1)
{
	uint32 val=0x80000000;
	static uint32 select_hold = 0;
	static uint32 vol_hold = 0;

	if (which1 != 0) return val;
	unsigned long joy = gp2x_joystick_read();

	if ((joy & (GP2X_VOL_UP | GP2X_VOL_DOWN)) == (GP2X_VOL_UP | GP2X_VOL_DOWN))
	{
		if ((gp2x_timer_read() - vol_hold) >= 1000)
		{ 		
			if (joy & GP2X_A) { PPU.BG_Forced ^= 1; vol_hold = gp2x_timer_read() - 500; }
			if (joy & GP2X_B) { PPU.BG_Forced ^= 2; vol_hold = gp2x_timer_read() - 500; }
			if (joy & GP2X_X) { PPU.BG_Forced ^= 4; vol_hold = gp2x_timer_read() - 500; }
			if (joy & GP2X_Y) { PPU.BG_Forced ^= 8; vol_hold = gp2x_timer_read() - 500; }  

			sprintf (displaytext, "BGs Active: %c %c %c %c", PPU.BG_Forced & 1 ? ' ' : '1',
				PPU.BG_Forced & 2 ? ' ' : '2', PPU.BG_Forced & 4 ? ' ' : '3', PPU.BG_Forced & 8 ? ' ' : '4');
			gettimeofday(&noticeMsgTime, 0);
		}
	}
	else if ((joy & (GP2X_L|GP2X_R|GP2X_START)) == (GP2X_L|GP2X_R|GP2X_START))
	{
		pause_to_menu = 1;
	}
	else if ((joy & (GP2X_VOL_UP | GP2X_VOL_DOWN)))
	{
		vol_hold = gp2x_timer_read();

		if ((joy & GP2X_VOL_UP))
		{
			if (sndvolL < 100) sndvolL ++; 
			if (sndvolR < 100) sndvolR ++;
			gp2x_set_sound_volume (sndvolL, sndvolR);
			if ((sndvolL == 100) && (sndvolR == 100))
			{
				strcpy (displaytext, "Vol Max");
				gettimeofday(&noticeMsgTime, 0);
			}
		}

		if ((joy & GP2X_VOL_DOWN))
		{
			if (sndvolL > 0) sndvolL --; 
			if (sndvolR > 0) sndvolR --; 
			gp2x_set_sound_volume (sndvolL, sndvolR);
			if ((sndvolL == 0) && (sndvolR == 0))
			{
				strcpy (displaytext, "Vol Min");
				gettimeofday(&noticeMsgTime, 0);
			}
		}
	}

	if (local_mr.joytype)
	{
		if (joy & GP2X_A) val |= SNES_Y_MASK;
		if (joy & GP2X_B) val |= SNES_A_MASK;
		if (joy & GP2X_X) val |= SNES_B_MASK;
		if (joy & GP2X_Y) val |= SNES_X_MASK;
	}
	else
	{
		if (joy & GP2X_A) val |= SNES_A_MASK;
		if (joy & GP2X_B) val |= SNES_B_MASK;
		if (joy & GP2X_X) val |= SNES_X_MASK;
		if (joy & GP2X_Y) val |= SNES_Y_MASK;
	}

	if (joy & GP2X_UP) val |= SNES_UP_MASK;
	if (joy & GP2X_DOWN) val |= SNES_DOWN_MASK;
	if (joy & GP2X_LEFT) val |= SNES_LEFT_MASK;
	if (joy & GP2X_RIGHT) val |= SNES_RIGHT_MASK;
	if (joy & GP2X_START) val |= SNES_START_MASK;
	if (joy & GP2X_L) val |= SNES_TL_MASK;
	if (joy & GP2X_R) val |= SNES_TR_MASK;
	if (joy & GP2X_SELECT) 
	{
		val |= SNES_SELECT_MASK;
		if (!select_hold)
		{
			select_hold = gp2x_timer_read();
		}
		else
		{
			if ((gp2x_timer_read() - select_hold) > 2000)
			{
				pause_to_menu = 1;
			}
		}
	}
	else
	{
		select_hold = 0;
	}

	return val;
}

extern "C" bool8 S9xReadMousePosition (int /* which1 */, int &/* x */, int & /* y */, uint32 & /* buttons */)
{
	S9xMessage (0,0,"read mouse");
	return (FALSE);
}

extern "C" bool8 S9xReadSuperScopePosition (int & /* x */, int & /* y */, uint32 & /* buttons */)
{
	S9xMessage (0,0,"read scope");
	return (FALSE);
}

extern "C" const char *S9xBasename (const char *f)
{
	const char *p;

	S9xMessage (0,0,"s9x base name");

	if ((p = strrchr (f, '/')) != NULL || (p = strrchr (f, '\\')) != NULL)
		return (p + 1);

	return (f);
}
   
#if 0
   void asmClearScreenFast16(uint8 *scr,uint32 starty,uint32 endy,uint32 back)
   {
   	unsigned short *screen = (unsigned short *) scr;
   	unsigned int sizey = ((endy - starty) + 1);
   	
   	screen += (320 * starty);
		while (sizey)
		{
			for (int i = 0; i < 4; i ++) 
			{
				*screen++ = back;	*screen++ = back; *screen++ = back; *screen++ = back; *screen++ = back;
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back; *screen++ = back;
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back; *screen++ = back;
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back; *screen++ = back;	
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back; *screen++ = back;
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back; *screen++ = back;
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back; *screen++ = back;	
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back; *screen++ = back;	
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back; *screen++ = back;
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back; *screen++ = back;
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back; *screen++ = back;
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back; *screen++ = back;
				*screen++ = back;	*screen++ = back;	*screen++ = back;	*screen++ = back;		
			}
			screen += 64;
			sizey--;
		}   	
   }
#endif

	/*void asmCSF_Screen (uint32 starty,uint32 endy,uint32 back)
	{
   	unsigned int sizey = ((endy - starty) + 1);
   	
   	if (!local_mr.wantscale)
   	{
			gp2x_blitter[0x4 >> 2] = (unsigned int) ((unsigned int)gp2x_video_RGB[0].screen16phys) + ((640 * (starty + 8)) + 64);
		}
		else
		{
			gp2x_blitter[0x4 >> 2] = (unsigned int) ((unsigned int)gp2x_video_RGB[0].screen16phys) + (640 * starty);
		}
	
		gp2x_blitter[0x0] = 1<<5; 											// Destination is 16 bpp
		gp2x_blitter[0x8 >> 2] = 640;										// Destination stride size in bytes
		gp2x_blitter[0x20 >> 2] = 1 << 5  | 2 << 3;					// Perform ROP with a 1bpp pattern
		gp2x_blitter[0x24 >> 2] = back;									// Setup the foreground and background colors for the pattern to be the same
		gp2x_blitter[0x28 >> 2] = back;									// so we don't actually have to upload a pattern image. 
		gp2x_blitter[0x2c >> 2] = (sizey << 16) | (256);			// Height and width to blit
		gp2x_blitter[0x30 >> 2] = (1<<8) | (1<<9 ) | 0xf0;			// Fill from top left to bottom right
		
		// Wait for blitter to be free, start it, and then wait for completion.
		// Throw in some nop's so we don't saturate the address bus with polling requests.
		while (gp2x_blitter[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
		gp2x_blitter[0x34 >> 2] = 1; 
		while (gp2x_blitter[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
	}
	
	void asmCSF_SubScreen (uint32 starty, uint32 endy)
	{
   	unsigned int sizey = ((endy - starty) + 1);
   	
		gp2x_blitter[0x4 >> 2] = 0x03000000;
		gp2x_blitter[0x0] = 1<<5; 											// Destination is 16 bpp
		gp2x_blitter[0x8 >> 2] = 640;										// Destination stride size in bytes
		gp2x_blitter[0x20 >> 2] = 1 << 5  | 2 << 3;					// Perform ROP with a 1bpp pattern
		gp2x_blitter[0x24 >> 2] = 0;										// Setup the foreground and background colors for the pattern to be the same
		gp2x_blitter[0x28 >> 2] = 0;										// so we don't actually have to upload a pattern image. 
		gp2x_blitter[0x2c >> 2] = (sizey << 16) | (256);			// Height and width to blit
		gp2x_blitter[0x30 >> 2] = (1<<8) | (1<<9 ) | 0xf0;			// Fill from top left to bottom right
		
		// Wait for blitter to be free, start it, and then wait for completion.
		// Throw in some nop's so we don't saturate the address bus with polling requests.
		while (gp2x_blitter[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
		gp2x_blitter[0x34 >> 2] = 1; 
		while (gp2x_blitter[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
	}
	
	void blitterCopySquidgeTransBuffer (void)
	{
   	if (!local_mr.wantscale)
   	{
			gp2x_blitter[0x4 >> 2] = (unsigned int) ((unsigned int)gp2x_video_RGB[0].screen16phys) + ((640 * 8) + 64);
		}
		else
		{
			gp2x_blitter[0x4 >> 2] = (unsigned int) ((unsigned int)gp2x_video_RGB[0].screen16phys);
		}
		gp2x_blitter[0x0] = 1<<5; 											// Destination is 16 bpp
		gp2x_blitter[0x8 >> 2] = 1280;									// Destination stride size in bytes
		
		gp2x_blitter[0xC >> 2] = 1 << 8 | 1 << 7 | 1 << 5; 		// if this works/doesnt, try with bit 7 clear.
		gp2x_blitter[0x10 >> 2] = 0x03061A80;							// Source address
		gp2x_blitter[0x14 >> 2] = 1280;									// Source stride
		
		gp2x_blitter[0x2c >> 2] = (224 << 16) | (256);				// Height and width to blit
		gp2x_blitter[0x30 >> 2] = (1<<11) | (1<<9) | (1<<8 ) | 0xcc;			// Fill from top left to bottom right
		
		// Wait for blitter to be free, start it, and then wait for completion.
		// Throw in some nop's so we don't saturate the address bus with polling requests.
		while (gp2x_blitter[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
		gp2x_blitter[0x34 >> 2] = 1; 
		while (gp2x_blitter[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
	}
	
	void blitterClearSquidgeTransBuffer (void)
	{
		gp2x_blitter[0x4 >> 2] = 0x03061A80;
		gp2x_blitter[0x0] = 1<<5; 											// Destination is 16 bpp
		gp2x_blitter[0x8 >> 2] = 640;										// Destination stride size in bytes
		gp2x_blitter[0x20 >> 2] = 1 << 5  | 2 << 3;					// Perform ROP with a 1bpp pattern
		gp2x_blitter[0x24 >> 2] = 0;										// Setup the foreground and background colors for the pattern to be the same
		gp2x_blitter[0x28 >> 2] = 0;										// so we don't actually have to upload a pattern image. 
		gp2x_blitter[0x2c >> 2] = (240 << 16) | (256);				// Height and width to blit
		gp2x_blitter[0x30 >> 2] = (1<<8) | (1<<9 ) | 0xf0;			// Fill from top left to bottom right
		
		// Wait for blitter to be free, start it, and then wait for completion.
		// Throw in some nop's so we don't saturate the address bus with polling requests.
		while (gp2x_blitter[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
		gp2x_blitter[0x34 >> 2] = 1; 
		while (gp2x_blitter[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
	}*/


void S9xAutoSaveSRAM (void)
{
	Memory.SaveSRAM (S9xGetFilename (".srm"));
}

bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
{
	return TRUE;
}

bool JustifierOffscreen(void)
{
	return false;
}

void JustifierButtons(uint32& justifiers)
{
}

#define SNES_SCANLINE_TIME (63.695e-6)
#define SNES_CLOCK_SPEED (3579545)

#define SNES_CLOCK_LEN (1.0 / SNES_CLOCK_SPEED)

//#define SNES_CYCLES_PER_SCANLINE ((uint32) ((SNES_SCANLINE_TIME / SNES_CLOCK_LEN) * 6 + 0.5))

void blankscreen (void)
{
	int x,y;	
	for (y = 0; y < 240; y ++)
		for (x = 0; x < 320; x ++)
			gp2x_video_RGB[0].screen16[(y*320)+x] = 0;
	gp2x_video_RGB_flip(0);
	for (y = 0; y < 240; y ++)
		for (x = 0; x < 320; x ++)
			gp2x_video_RGB[0].screen16[(y*320)+x] = 0;
	gp2x_video_RGB_flip(0);
}

void drawbackground (void)
{
	int i=0;
	// draw the background image onto both buffers if possible
	if (!gamebg) return;
	if (local_mr.wantscale) return;
	
	unsigned short *dstpix;
	unsigned char *pix;
	
	for (i=0;i<4;i++)
	{
	dstpix = gp2x_video_RGB[0].screen16;
	pix = gamebg->getPixels();
	
	for (int yy = 0; yy < 240; yy ++)
      for (int xx = 0; xx < 320; xx ++)
      {
         unsigned char r, g, b;
         unsigned short p;

         r = pix[(yy*(320*3))+(xx*3)+0];
         g = pix[(yy*(320*3))+(xx*3)+1];
         b = pix[(yy*(320*3))+(xx*3)+2];

         r /= 8;
         g /= 4;
         b /= 8;

         r &= 31;
         g &= 63;
         b &= 31;

         p = (b << 11 ) | (g << 5) | (r);
         *dstpix++ = p;
      }
		gp2x_video_RGB_flip(0);
      }
}


void copymenuopts (struct menureturn *mr)
{
	static int gp2x_clock_old = 200, ram_timings_set = 0, gamma_old = 100;

	if (!mr) return;
	memcpy (&local_mr, mr, sizeof(menureturn));

	//printf("cfg: apuenable: %i\n", local_mr.apuenable);
	
	// Stuff we need to pass to the emu core.	
//	gp32_fastsprite = local_mr.fastsprites;
	squidgetranshack = local_mr.squidgetranshack;

	// stuff that may need to be changed
	if (local_mr.gp2x_clock != gp2x_clock_old && local_mr.gp2x_clock >= 10)
	{
		printf("changing clock to %i...", local_mr.gp2x_clock); fflush(stdout);
		set_FCLK(local_mr.gp2x_clock);
		gp2x_clock_old = local_mr.gp2x_clock;
		printf(" done\n");
	}

	if (local_mr.ramtimings && !ram_timings_set)
	{
		// craigix: --trc 6 --tras 4 --twr 1 --tmrd 1 --trfc 1 --trp 2 --trcd 2
		printf("setting RAM timings.. "); fflush(stdout);
		set_RAM_Timings(6, 4, 1, 1, 1, 2, 2);
		ram_timings_set = 1;
		printf("done.\n"); fflush(stdout);
	}

	if (local_mr.gamma != gamma_old && local_mr.gamma >= 10)
	{
		set_gamma(local_mr.gamma);
		gamma_old = local_mr.gamma;
		printf("updated gamma to %i\n", local_mr.gamma);
	}
}


// simple single-threaded method worked well for PicoDrive,
// but here some strange slowdowns happened, so I use this
void *soundthread_f (void *a)
{
	int sval = 0;

	while (1)
	{
		sem_wait(&samplebuffer_sem);
		// be sure that sem count is not too high
		sem_getvalue(&samplebuffer_sem, &sval);
		while (sval > 0) {
			sem_wait(&samplebuffer_sem);
			sem_getvalue(&samplebuffer_sem, &sval);
		}

		gp2x_sound_write(samplebuffer, samplebuffer_bytes); // also does SOUND_PCM_SYNC
	}
	
	return NULL;
}


// save state file I/O
int  (*statef_open)(const char *fname, const char *mode);
int  (*statef_read)(void *p, int l);
int  (*statef_write)(void *p, int l);
void (*statef_close)();
static FILE  *state_file = 0;

int state_unc_open(const char *fname, const char *mode)
{
	state_file = fopen(fname, mode);
	return (int) state_file;
}

int state_unc_read(void *p, int l)
{
	return fread(p, 1, l, state_file);
}

int state_unc_write(void *p, int l)
{
	return fwrite(p, 1, l, state_file);
}

void state_unc_close()
{
	fclose(state_file);
}

static int loadROM(const char *name)
{
	int ret;
	const char *cfgfile;
	FILE *cfgf;

	blankscreen();
	gp2x_printf(NULL, 0, 0,"Attempting to load rom:\n%s", name);
	gp2x_video_RGB_flip(0);
	gp2x_printf(NULL, 0, 0,"Attempting to load rom:\n%s", name);
	if ((ret = Memory.LoadROM(name)))
	{
		printf("Memory.LoadROM(%s) failed with %i\n", name, ret);
		return ret;
	}
	gp2x_printf(NULL, 0, 0, "Attempting to load SRAM:\n%s", S9xGetFilename (".srm"));
	gp2x_video_RGB_flip(0);
	Memory.LoadSRAM (S9xGetFilename (".srm"));

	// try to load ROM specific config
	cfgfile = S9xGetFilename(".cfg");
	cfgf = fopen(cfgfile, "rb");
	if (cfgf) {
		struct menureturn *mr;
		fclose(cfgf);
		printf("loading cfg: %s\n", cfgfile);
		mr = skipmenu(cfgfile);
		copymenuopts(mr);
	} else {
		printf("game cfg (%s) not found.\n", cfgfile);
	}

	return 0;
}

static void EmuGo (void)
{
	unsigned int i = 0;
	unsigned int romrunning = 0;

	ZeroMemory (&Settings, sizeof (Settings));

	Settings.JoystickEnabled = FALSE;
//	Settings.SoundPlaybackRate = 4; // unused
	Settings.Stereo = FALSE;
//	Settings.SoundBufferSize = 1024; // unused
	Settings.CyclesPercentage = 100;
	Settings.DisableSoundEcho = FALSE;
	Settings.APUEnabled = FALSE;
	Settings.H_Max = SNES_CYCLES_PER_SCANLINE;
	Settings.SkipFrames = AUTO_FRAMERATE;
	Settings.ShutdownMaster = TRUE;
	Settings.FrameTimePAL = 20;
	Settings.FrameTimeNTSC = 17;
	Settings.FrameTime = Settings.FrameTimeNTSC;
	Settings.DisableSampleCaching = FALSE;
	Settings.DisableMasterVolume = FALSE;
	Settings.Mouse = FALSE;
	Settings.SuperScope = FALSE;
	Settings.MultiPlayer5 = FALSE;
	//	Settings.ControllerOption = SNES_MULTIPLAYER5;
	Settings.ControllerOption = 0;

	if(gp32_8bitmode)
	{
		Settings.ForceTransparency = FALSE;
		Settings.Transparency = FALSE;
		Settings.SixteenBit = FALSE;
	}
	else
	{
		Settings.ForceTransparency = FALSE;
		Settings.Transparency = FALSE;
		Settings.SixteenBit = TRUE;
	}

	Settings.SupportHiRes = FALSE;
	Settings.NetPlay = FALSE;
	Settings.ServerName [0] = 0;
	Settings.ThreadSound = TRUE;
	Settings.AutoSaveDelay = 30;
	Settings.ApplyCheats = FALSE;
	Settings.TurboMode = FALSE;
	Settings.TurboSkipFrames = 15;
	Settings.SoundSync = FALSE;
	//Settings.NoPatch = true;

	vrambuffer = (uint8 *) malloc (640 * 480 * 2);
	memset (vrambuffer, 0, 640*480*2);
	GFX.Screen = vrambuffer+32*240*2;
	GFX.RealPitch = GFX.Pitch = 240 * 2;
	GFX.SubScreen = gp2x_map_upper_ram(0); //(uint8 *)malloc(640 * 480 * 2);
	GFX.ZBuffer = gp2x_map_upper_ram(200000); //(uint8 *)malloc(640 * 480 * 2);
	GFX.SquidgeTransBuffer = gp2x_map_upper_ram(400000);
	GFX.SubZBuffer = (uint8 *)malloc(640 * 480 * 2);
	GFX.Delta = (GFX.SubScreen - GFX.Screen) >> 1;
	GFX.PPL = GFX.Pitch >> 1;
	GFX.PPLx2 = GFX.Pitch;
	GFX.ZPitch = GFX.Pitch >> 1;

	if (Settings.ForceNoTransparency)
	Settings.Transparency = FALSE;

	if (Settings.Transparency)
	Settings.SixteenBit = TRUE;

	Settings.HBlankStart = (256 * Settings.H_Max) / SNES_HCOUNTER_MAX;

	if (!Memory.Init () || !S9xInitAPU())
		erk();

	S9xInitSound();

	//S9xSetRenderPixelFormat (RGB565);
	S9xSetSoundMute (TRUE);

	if (!S9xGraphicsInit ())
		erk();

	// saves
	statef_open  = state_unc_open;
	statef_read  = state_unc_read;
	statef_write = state_unc_write;
	statef_close = state_unc_close;

	sem_init(&samplebuffer_sem, 0, 0);

	try
	{
		gamebg = new hm::Image ("skin/gamebackground.png");
	}
	catch (...)
	{
		gamebg = NULL;
	}

	while (1)
	{
		struct menureturn *mr;
		int sound_len = 0, ret;
		int snd_excess_add = 0, snd_excess_cnt = 0;

		if (!cmdline_rom[0])
		{
			S9xSetSoundMute (TRUE);
			gp2x_video_RGB_setscaling (320, 240);
			if (romrunning)
			     mr = domenu(0, S9xGetFilename(".cfg"));
			else mr = domenu(1, NULL);
			copymenuopts(mr);

			// what was selected?
			if (!mr->name[0])
			{
				// nothing! user escaped the menu system. if a rom is loaded go back to game, otherwise exit to os. 
				if (!romrunning) break;
			}

			// stop anyone using "return to game" when there's no rom loaded
			if (!strcasecmp(mr->name, "returntogame"))
			{
				if (!romrunning) continue;
			}

			if (!strcasecmp(mr->name, "loadrom.file"))
			{
				// user wants to load a rom
				romrunning = 0;
				ret = loadROM(mr->value);
				if (ret) continue;
			}

			if (!strcasecmp(mr->name, "resetaction"))
			{
				// reset snes
				if (!romrunning) continue;
				S9xReset();
			}

			if (!strcasecmp(mr->name, "exitaction"))
			{
				// exit emu
				break;
			}

			if (!strcasecmp(mr->name, "loadstate"))
			{
				if (!romrunning) continue;
				S9xUnfreezeGame (S9xGetFilename (".state"));
			}

			if (!strcasecmp(mr->name, "savestate"))
			{
				if (!romrunning) continue;
				S9xFreezeGame (S9xGetFilename (".state"));
				sync();
			}
		}
		else
		{
			mr = skipmenu("ssnes.cfg");
			copymenuopts(mr);

			romrunning = 0;
			ret = loadROM(cmdline_rom);
			if (ret)
			{
				printf("exitting\n"); fflush(stdout);
				break;
			}
			cmdline_rom[0] = 0;
		}

		// any change in configuration?
		CPU.APU_APUExecuting = Settings.APUEnabled = local_mr.apuenable;

		if(gp32_8bitmode)
		{
			gp2x_init (1000, 8, 44100, 16, 1, 60, 1);
		}
		else
		{
			blankscreen();
			drawbackground();
		}
		oldscaley = 0;

		if (Memory.ROMFramesPerSecond == 0) break;

		// timing
		int target_fps = Memory.ROMFramesPerSecond;
		struct timeval tval;
		gettimeofday(&tval, 0);
		target_frametime = 1000000 / target_fps;
		frames_shown = frames_done = tval.tv_usec/target_frametime;

		if (soundthread_p == (pthread_t) -1) 
			pthread_create(&soundthread_p, NULL, &soundthread_f, NULL);

		// sound?
		if ((local_mr.apuenable & 1) && !local_mr.apumute)
		{
			so.stereo = Settings.Stereo = local_mr.stereo;
			so.playback_rate = local_mr.audio_rate;
			if (so.stereo) so.stereo = 1;
			S9xSetPlaybackRate(so.playback_rate);
			S9xSetSoundMute (FALSE);

			sound_len = so.playback_rate / Memory.ROMFramesPerSecond;
			if (so.stereo) sound_len<<=1;
			snd_excess_cnt = 0;
			snd_excess_add = ((so.playback_rate - (so.playback_rate/target_fps)*target_fps)<<16) / Memory.ROMFramesPerSecond; // int math
			printf("starting audio: %i len: %i (ex: %04x) stereo: %i, pal: %i\n",
					so.playback_rate, sound_len, snd_excess_add, so.stereo, Settings.PAL);
			gp2x_start_sound(so.playback_rate, 16, so.stereo);
		}
		else
		{
			S9xSetSoundMute (TRUE);
		}

		// bug: sometimes slowdowns or no display just after exitting menu. Why?
		//usleep(200*1000);
		usleep(1);

		while (1)
		{
			S9xMainLoop();
			if (sound_len) {
				int sound_add = 0;
				snd_excess_cnt += snd_excess_add;
				if (snd_excess_cnt >= 0x10000) {
					snd_excess_cnt -= 0x10000;
					sound_add = so.stereo + 1;
				}

				S9xMixSamplesO(samplebuffer, sound_len+sound_add, 0);
				//gp2x_sound_write(samplebuffer, (sound_len+sound_add)<<1);
				samplebuffer_bytes = (sound_len+sound_add)<<1; // might have sync problems
				sem_post(&samplebuffer_sem);
			}

			if (pause_to_menu) break;
		}

		Memory.SaveSRAM (S9xGetFilename (".srm"));

		if(gp32_8bitmode)
		{
			gp2x_init (1000, 16, 44100, 16, 1, 60, 1);
		}
		romrunning = 1;
		pause_to_menu = 0;
	}

	if (gamebg)
	{
		delete gamebg;
		gamebg = NULL;
	}
}

//---------------------------------------------------------------------------

static void checkargs (char *arg)
{
	if (!strcasecmp(arg, "nrtm"))
		chain_menu = 0;
	if (!strncasecmp(arg, "load=", 5)) {
		strncpy (cmdline_rom, &arg[5], 255);
	}
}


static int devbatt; 
static int battvolts;

void initbatt (void)
{
	devbatt = open ("/dev/batt", O_RDONLY);
	if (devbatt == -1)
	{
		printf ("Unable to open /dev/batt\n");
		return;
	}
}

void deinitbatt (void)
{
	if (devbatt != -1)
	{
		close (devbatt);
		devbatt = -1;
	}
}

static int battConvert (int battval)
{
	int v;

	if (battval > 1016) v = 37;
	else if (battval > 974) v = 33;
	else if (battval > 943) v = 32;
	else if (battval > 915) v = 31;
	else if (battval > 896) v = 30;
	else if (battval > 837) v = 29;
	else if (battval > 815) v = 28;
	else if (battval > 788) v = 27;
	else if (battval > 745) v = 26;
	else if (battval > 708) v = 25;
	else if (battval > 678) v = 24;
	else if (battval > 649) v = 23;
	else if (battval > 605) v = 22;
	else if (battval > 573) v = 21;
	else if (battval > 534) v = 20;
	else if (battval > 496) v = 19;
	else if (battval > 448) v = 18;
	else v = 17;	

	return v;
}

void checkbatt (void)
{
	if (devbatt == -1)
	{
		battvolts = 27;
	}
	else
	{
		static int firsttime = 1;
		static int battlevel = 0;
		static int i = 0;
		unsigned short cbv;
		
		if (read (devbatt, &cbv, 2) == 2)
			battlevel += cbv;
		
		if (firsttime)
		{
			battvolts = battConvert (battlevel);
			battlevel = 0;
			firsttime = 0;
		}
		else
		{
			i++;
			if (i >= 50)
			{
				battlevel /= 50;
				battvolts = battConvert (battlevel);
				battlevel = 0;		
				i = 0;		
			}
		}
	}
}

// Used by the menu system
int battChargeLevel (void)
{
	int cl;
	
	if (battvolts >= 27)
		cl = 100;
	else if (battvolts >= 26)
		cl = 70;
	else if (battvolts >= 25)
		cl = 40;
	else if (battvolts >= 24)
		cl = 10;
	else
		cl = 0;
	
	return cl;
}


int main(int argc, char **argv)
{
	g_argv = argv;

	cmdline_rom[0] = 0;
	if (argc > 1)
	{
		checkargs (argv[1]);
	}
	if (argc > 2)
	{
		checkargs (argv[2]);
	}

	gp2x_init (1000, 16, 44100, 16, 1, 60, 1);
	cpuctrl_init();
	//gp2x_video_logo_enable(1);
	gp2x_printf(NULL, 0, 0,"%s\n%s\nCommand line: %d args\n",EMUVERSION,MINILIB_VERSION, argc); 
	gp2x_video_RGB_flip(0);
	gp2x_printf(NULL, 0, 0,"%s\n%s\nCommand line: %d args\n",EMUVERSION,MINILIB_VERSION, argc); 

	//initaudio();	
	initbatt();

	EmuGo();

	cpuctrl_deinit();
	deinitbatt();
	//deinitaudio();

	return 0;
}
