#include <string>
#include "surface.h"
#include "actionmenu.h"
#include "system.h"

using namespace std;

namespace hm
{
	ActionMenu::ActionMenu(string title,Surface *surface) :
		Menu(title,surface)
	{

		width = 240;
		height = 120;
		confirm = false;
		down = false;
		selectedButton = 0;

	}

	void ActionMenu::draw()
   {
		int x = this->x;
		int y = this->y;

		Surface *surface = getSurface();

		surface->clearClipRect();

		// draw background
		if (background!=NULL)
			surface->drawImage(background,0,0);

		// center the window to the surface
      if ((x==-1)&&(y==-1))
      {
		   x += (surface->getWidth() - width) / 2;
		   y += (surface->getHeight() - height) / 2;
      }

		// draw main window
		surface->drawWindow(x,y,width,height,"Are you sure?");

		// draw alert icon
		surface->drawAlertIcon(x+windowPadding,y+surface->getHeaderHeight()+windowPadding);
		
		// TODO: the 20 is just a guess, not the real width of the icon
		surface->drawText(confirmText,x+windowPadding+40,y+surface->getHeaderHeight()+windowPadding);		

		// draw buttons
		surface->drawButton(
			"Ok",
			x+width/2 - surface->getButtonWidth() - windowPadding,
			y+height-surface->getButtonHeight() - windowPadding,
			(selectedButton==0)&&down,
			selectedButton==0
			);
		surface->drawButton(
			"Cancel",
			x+width/2 + windowPadding,
			y+height-surface->getButtonHeight() - windowPadding,
			(selectedButton==1)&&down,
			selectedButton==1
			);

		surface->flip();

	}

	int ActionMenu::enter()
   {
      int res = 0;

		// fire enter event
		if (actionHandler!=NULL)
			actionHandler->enter(this);

		Surface *surface = getSurface();

		if (confirm)
      {

			background = surface->toImage();

			selectedButton = 0;
			draw();

			running=true;
			while (running)
			{
				Event e = msystem.pollEvent();
				if (e.type==EVENT_QUIT) running = false;

				if (e.type==EVENT_KEYUP)
            {
					if (e.key==KEY_REJECT)
						running=false;

					if (e.key==KEY_ACCEPT)
               {
						down = false;
						draw();
						running=false;
						if (selectedButton==0)
                  {
							if (actionHandler!=NULL)
                     {
								if (actionHandler->action(this))
                        {
                           res = 1;
                           break;
                        }
                     }
						}
					}
				}

				if (e.type==EVENT_KEYDOWN)
            {

					if (e.key==KEY_LEFT) selectedButton = 0;
					if (e.key==KEY_RIGHT) selectedButton = 1;
					if (e.key==KEY_ACCEPT) down = true;

					draw();
				}

			}

			delete(background);
			background = NULL;
		}
      else
      {

			if (actionHandler!=NULL)
         {
				if (actionHandler->action(this))
            {
               res = 1;
            }
         }
		}
      return res;
	}

	void ActionMenu::setConfirm(bool conf,string text)
   {
		confirm = conf;
		confirmText = text;
	}

};
