#ifndef __configfile_h__
#define __configfile_h__

#include <string>
#include <vector>

using namespace std;

namespace hm {

	/** 
		The ConfigFile is used to save the state of a menu structure to a file. It is basically a collection of name,value
		tuples. 
	*/
	class ConfigFile {

	public:

		/** adds a name,value pair to the collection */
		void addVariable(string name,string value);

		/** saves all name,value tuples to a file
			\param filename name of the file to save to
			\throws <string> if the file can't be opened
		*/
		void save(string filename);

		/** loads a bunch of name,value tuples from a file. If the file can't be opened, it will return empty handed.
			\param filename name of the file to load from
		*/
		void load(string filename);

		/** make it forget all tuples.
		*/
		void clear();

		/** Returns the value corresponding to a given name.
			\param name of the requested value
			\return <string> requested value or "" if not found
		*/
		string getValueByName(string name);

		/** gets the value corresponding to a given name, then runs it through atoi and returns the result
			\param name of the requested value
			\return <int> requested value or 0 if not found/not an integer
		*/
		int getIntValueByName(string name);

	private:

		typedef struct {
			string name;
			string value;
		} Variable;

		typedef std::vector<Variable> VariableList;

		VariableList variables;


	};

};

#endif
