#include "font.h"
#include "image.h"

namespace hm
{

	const static char* alphabet = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'\"[]{}()!@#$%^&*~`-_=+.,<>/?\\;:";

	// TODO: error handling
	Font::Font(Image* fontImage)
	{

		unsigned char *fontPixels = fontImage->getPixels();
		int width = fontImage->getWidth();
		height = fontImage->getHeight();

		// allocate memory for all the individual letter images
		letters = new Image*[strlen(alphabet)];

		// assume first line is not empty
		bool wasEmpty = true;

		// scan for individual letters
		int start;
		nrLetters=0;
		for (int i=0; i<width; i++)
		{

			// is this an empty line ?
			bool isEmpty = true;
			for (int j=0; j<height; j++)
			{
				isEmpty &= (fontPixels[(j*width+i)*3+0]==255);
				isEmpty &= (fontPixels[(j*width+i)*3+1]==0);
				isEmpty &= (fontPixels[(j*width+i)*3+2]==255);
			}

			// is there a transition here from empty->nonempty ?
			if (wasEmpty&&!isEmpty)
				start = i;

			// is there a transition here from nonempty->empty ?
			if (!wasEmpty&&isEmpty)
			{
				letters[nrLetters] = fontImage->subImage(start,0,i-start,height);
				nrLetters++;
			}

			wasEmpty = isEmpty;
		}

		// default
		letterSpacing = 1;
		wordSpacing = 4;

		// delete font image
		delete(fontImage);
	}

	Font::~Font()
	{

		// free letter images
		for (int i=0; i<nrLetters; i++)
			delete(letters[i]);

		delete[] letters;

	}

	Image* Font::getLetter(char c)
	{
		for (unsigned int i=0; i<strlen(alphabet); i++)
			if (alphabet[i]==c) return(letters[i]);
		return(NULL);
	}

	void Font::setLetterSpacing(int space)
	{
		letterSpacing = space;
	}

	void Font::setWordSpacing(int space)
	{
		wordSpacing = space;
	}

	int Font::getLetterSpacing()
	{
		return(letterSpacing);
	}

	int Font::getWordSpacing()
	{
		return(wordSpacing);
	}

	int Font::getHeight()
	{
		return(height);
	}

	int Font::getStringWidth(string text)
	{
		char *txt = (char*)text.c_str();
		int x = 0;
		for (int i=0; i<strlen(txt); i++)
		{
			Image *letter = getLetter(txt[i]);
			if (letter!=NULL)
			{
				x += letter->getWidth() + getLetterSpacing();
			}
			else
			{
				x += getWordSpacing();
			}
		}
		return(x);
	}


}
