#include <string>
#include "surface.h"
#include "launchermenu.h"
#include "system.h"
#include "math.h"
using namespace std;

namespace hm
{

	LauncherMenu::LauncherMenu(string title,Surface *surface) :
		Menu(title,surface)
	{
		setValue("");
		setWindowSize(280,200);
		selected=0;
	}

	void LauncherMenu::draw()
   {

		int x = this->x;
		int y = this->y;

		Surface *surface = getSurface();

		surface->drawImage(background,0,0);

		// center the window to the surface
      if ((x==-1) && (y==-1))
      {
		   x = (surface->getWidth() - width) / 2;
		   y = (surface->getHeight() - height) / 2;
      }

		surface->drawWindow(x,y,width,height,title);

		x += windowPadding;
		y += windowPadding+surface->getHeaderHeight();

		drawList(
			x+windowPadding,
			y,
			width-windowPadding*3,
			height-windowPadding*2-surface->getHeaderHeight());

		surface->flip();

	}

	int LauncherMenu::enter()
   {
      int res = 0;

		// fire enter event
		if (actionHandler!=NULL)
			actionHandler->enter(this);

		background = getSurface()->toImage();

		draw();
		scroll = 0;

		running=true;
		while (running)
		{
			Event e = msystem.pollEvent();
			if (e.type==EVENT_QUIT) running = false;

			if (e.type==EVENT_KEYUP)
         {
				if (e.key==KEY_REJECT)
					running=false;

				if (e.key==KEY_ACCEPT)
            {
					if (children.size()>selected)
               {
						res = children.at(selected)->enter();
                  if (res)
                     break;
               }
					draw();
				}
			}

			if (e.type==EVENT_KEYDOWN)
         {
				if ((e.key==KEY_DOWN)&&(selected<children.size()-1))
					selected++;

				if ((e.key==KEY_UP)&&(selected>0))
					selected--;

				draw();
			}
		}

		delete(background);
		background = NULL;

      return res;
	}

};
