#include "hypermenu.h"
#include <string>
#include <sstream>
#include <iostream>

#include "menusys.h"

using namespace hm;
using namespace std;

static ConfigFile *configFile;
static Menu *menu;
static struct menureturn mr;
static FileMenu* loadrommenu;
static const char *gameConfigName = NULL;

class myActionHandler : public ActionHandler
{

	int action(Menu *source);
	void change(Menu *source);
	void enter(Menu *source);

};

int myActionHandler::action(Menu *source)
{
	if (source->getName() == "saveoptions")
	{
		configFile->clear();
		menu->addValues(configFile);
		configFile->save("ssnes.cfg");
		return 0;
	}
	else if (source->getName() == "saveoptions_game")
	{
		configFile->clear();
		menu->addValues(configFile);
		configFile->save(gameConfigName);
		return 0;
	}

	strncpy (mr.title, source->getTitle().c_str(), 255);
	strncpy (mr.name, source->getName().c_str(), 255);
	strncpy (mr.value, source->getValue().c_str(), 255);
	// return false to carry on with menu, 1 to exit menu
	return 1;
}

void myActionHandler::change(Menu *source)
{
	printf("Caught change event: [%s] = (%s,%s) \n",source->getTitle().c_str(),source->getName().c_str(),source->getValue().c_str());
   if (source->getName() == "absromdir")
   {
      if (source->getValue() == "true")
         loadrommenu->setValue("/mnt/sd/roms/snes");
      else
         loadrommenu->setValue("roms");
   }
}

void myActionHandler::enter(Menu *source)
{

	printf("Caught enter event: [%s] = (%s,%s) \n",source->getTitle().c_str(),source->getName().c_str(),source->getValue().c_str());

}

void parseconfig (void)
{
	string temp;
	char tmp[16];
	
   if (configFile->getValueByName("wantscale") == "true")
      mr.wantscale = 1;
   else
      mr.wantscale = 0;

   if (configFile->getValueByName("fpsdisplay") == "true")
      mr.wantfps = 1;
   else
      mr.wantfps = 0;

   if (configFile->getValueByName("absromdir") == "true")
      mr.romdir = 1;
   else
      mr.romdir = 0;

   if (configFile->getValueByName("snesjoytype") == "true")
      mr.joytype = 1;
   else
      mr.joytype = 0;
      
   if (configFile->getValueByName("apuenable") == "true")
   	mr.apuenable = 1;
   else
   	mr.apuenable = 0;

   if (configFile->getValueByName("apumute") == "true")
   	mr.apumute = 1;
   else
   	mr.apumute = 0;

   if (configFile->getValueByName("spc700hack") == "true" && (mr.apuenable & 1))
   	mr.apuenable |=  2;
   else
   	mr.apuenable &= ~2;

   temp = configFile->getValueByName("audio_quality");
   sscanf(temp.c_str(), "%iHz%6s", &mr.audio_rate, tmp);
   if (mr.audio_rate > 44100 || mr.audio_rate < 8000) mr.audio_rate = 16000;
   mr.stereo = (strcmp(tmp, "Stereo") == 0) ? 1 : 0;

   if (configFile->getValueByName("vsync") == "false")
      mr.vsyncenable = 0;
   else
      mr.vsyncenable = 1;
      
//	if (configFile->getValueByName("fastsprites") == "true")
//		mr.fastsprites = 0;
//	else
//		mr.fastsprites = 0;
		
	if (configFile->getValueByName("squidgetranshack") == "true")
		mr.squidgetranshack = 1;
	else
		mr.squidgetranshack = 0;
   	  	
   temp = configFile->getValueByName("lowbattval");
	if (temp == "2.7V") mr.lowbattval = 27;
	else if (temp == "2.6V") mr.lowbattval = 26;
	else if (temp == "2.5V") mr.lowbattval = 25;
	else if (temp == "2.4V") mr.lowbattval = 24;
	else if (temp == "2.3V") mr.lowbattval = 23;
	else if (temp == "2.2V") mr.lowbattval = 22;
	else if (temp == "2.1V") mr.lowbattval = 21;
	else if (temp == "2.0V") mr.lowbattval = 20;
	else mr.lowbattval = 0;

	if (configFile->getValueByName("fskip") != "Auto")
		 mr.fskip = configFile->getIntValueByName("fskip");
	else mr.fskip = -1;

	// notaz
	mr.gp2x_clock = configFile->getIntValueByName("gp2x_clock");

	if (configFile->getValueByName("ramtimings") == "true")
		mr.ramtimings = 1;
	else
		mr.ramtimings = 0;

	mr.gamma = configFile->getIntValueByName("gamma");
}

void applyDefaultConfig (ConfigFile *cf)
{
	cf->clear();
	cf->addVariable ("loadrom", "roms");
	cf->addVariable ("wantscale", "false");
	cf->addVariable ("fpsdisplay", "false");
	cf->addVariable ("vsync", "false");
	cf->addVariable ("absromdir","false");
	cf->addVariable ("snesjoytype","false");
	cf->addVariable ("apuenable","false");
	cf->addVariable ("apumute","true");
	cf->addVariable ("audio_quality","16000HzStereo");
	cf->addVariable ("spc700hack","false");
	cf->addVariable ("fskip","Auto");
//	cf->addVariable ("fastsprites","false");
	cf->addVariable ("squidgetranshack", "false");
	cf->addVariable ("lowbattval", "2.4V");
	// notaz
	cf->addVariable ("gp2x_clock","200");
	cf->addVariable ("ramtimings","false");
	cf->addVariable ("gamma","100");
}

static Image *images[128];
static int image_c = 0;

static Image *new_image(string fname)
{
	if (image_c > 127)
	{
		printf("new_image: img overflow");
		image_c = 0;
	}
	return (images[image_c++] = new Image(fname));
}

static void delete_images(void)
{
	for (image_c--; image_c >= 0; image_c--)
		delete images[image_c];
	image_c = 0;
}

struct menureturn *domenu(int sel, const char *gameCfgName)
{
	Surface *surface = 0;
	myActionHandler *actionHandler = 0;

	gameConfigName = gameCfgName;
	image_c = 0;

	mr.title[0] = 0;
	mr.name[0] = 0;
	mr.value[0] = 0;

	// initialize system
	msystem.initSystem();

	// try initializing a drawing surface
	try
	{
		surface = new Surface(320,240);
		Image* background = new_image("skin/background.png");
		surface->drawImage(background,0,0);
		surface->flip();
	} catch (string message) {
		// probably an image that couldn't be found
		cout << "error initializing surface: " << message << endl;
		exit(1);
	}

	// build the menu tree
	try
	{
		// create a new instance ouf our custom action handler
		actionHandler = new myActionHandler();

		ConfigFile *skinConfig = new ConfigFile();
		skinConfig->load("skin/skin.cfg");

		int wxp, wyp, wxs, wys, mt, i;
		mt = skinConfig->getIntValueByName("menutype");

		// initialize our menu tree root
		if (mt == 0)
		{
			wxs = 260;
			wys = 220;
			wxp = -1;
			wyp = -1;
			menu = new Menu("SquidgeSNES",surface);
		}
		else
		{
			wxp = skinConfig->getIntValueByName("menuxpos");
			wyp = skinConfig->getIntValueByName("menuypos");
			wxs = skinConfig->getIntValueByName("menuxsize");
			wys = skinConfig->getIntValueByName("menuysize");

			menu = new SquidgeMenu ("SquidgeSNES", surface);
		}

		menu->setWindowSize(wxs,wys); // 260x160 for new menu
		menu->setWindowPosition(wxp, wyp); //(320-260)/2,44

		delete(skinConfig);

		// return to game
		ActionMenu *rtgaction = new ActionMenu("Return to game");
		rtgaction->setActionHandler(actionHandler);
		rtgaction->setConfirm(false,"");
		rtgaction->setName("returntogame");
		rtgaction->setIcon(new_image("skin/icons/continue.png"));
		rtgaction->setSecondaryIcon(new_image("skin/icons/loadrom.png"));
		menu->appendChild(rtgaction);

		// load rom
		loadrommenu = new FileMenu("Load ROM", surface);
		loadrommenu->setIcon(new_image("skin/icons/loadrom.png"));
		loadrommenu->setActionHandler(actionHandler);
		loadrommenu->setName("loadrom");
		loadrommenu->setValue ("roms");
		loadrommenu->addFileExtension(".smc", new_image("skin/file.png"));
		loadrommenu->addFileExtension(".zip", new_image("skin/file.png"));
		menu->appendChild(loadrommenu);

		// Load state
		ActionMenu *loadstate = new ActionMenu("Load state");
		loadstate->setActionHandler(actionHandler);
		loadstate->setName("loadstate");
		loadstate->setIcon(new_image("skin/icons/loadstate.png"));
		menu->appendChild(loadstate);

		// Save state
		ActionMenu *savestate = new ActionMenu("Save state");
		savestate->setActionHandler(actionHandler);
		savestate->setName("savestate");
		savestate->setIcon(new_image("skin/icons/savestate.png"));
		menu->appendChild(savestate);

		// Reset snes
		ActionMenu *rsaction = new ActionMenu("Reset SNES");
		rsaction->setActionHandler(actionHandler);
		rsaction->setConfirm(true,"ROM will be restarted");
		rsaction->setName("resetaction");
		rsaction->setIcon(new_image("skin/icons/resetsnes.png"));
		menu->appendChild(rsaction);

		//
		// Create the configuration menu
		//

		Menu* configmenu = new Menu("Settings");
		configmenu->setIcon(new_image("skin/icons/settings.png"));
		configmenu->setWindowSize(200,200);
		menu->appendChild(configmenu);

		//
		// Configuration menu: Graphics
		//

		CheckboxMenu *wantscale = new CheckboxMenu("Use full screen scaling");
		wantscale->setName("wantscale");
		configmenu->appendChild(wantscale);

		CheckboxMenu *fpsdisplay = new CheckboxMenu("Show FPS");
		fpsdisplay->setName("fpsdisplay");
		configmenu->appendChild(fpsdisplay);

		CheckboxMenu *vsync = new CheckboxMenu ("Enable vsync");
		vsync->setName("vsync");
		configmenu->appendChild(vsync);

		CheckboxMenu *transhack = new CheckboxMenu ("Transhack-A V0.1");
		transhack->setName("squidgetranshack");
		configmenu->appendChild(transhack);

		SelectMenu *fskip = new SelectMenu("Frameskip");
		fskip->setName("fskip");
		fskip->addOption("Auto");
		for (int i = 0; i < 10; i ++)
		{
			char str[16];

			sprintf (str, "%d", i);
			fskip->addOption(str);
		}
		configmenu->appendChild(fskip);
		fskip->setValue("Auto");

		// notaz
		SelectMenu *gp2x_clock = new SelectMenu("GP2X CPU clock");
		gp2x_clock->setName("gp2x_clock");
		for (i = 120; i < 240; i+=10)
		{
			char str[16];

			sprintf (str, "%d", i);
			gp2x_clock->addOption(str);
		}
		for (i = 240; i < 320; i+=5)
		{
			char str[16];

			sprintf (str, "%d", i);
			gp2x_clock->addOption(str);
		}
		configmenu->appendChild(gp2x_clock);
		gp2x_clock->setValue("200");

		CheckboxMenu *ramtimings = new CheckboxMenu ("craigix's RAM timings");
		ramtimings->setName("ramtimings");
		configmenu->appendChild(ramtimings);

		SelectMenu *gamma = new SelectMenu("Gamma correction");
		gamma->setName("gamma");
		//	gamma->addOption("100"); // this menu requires default options to be first?
		// bug?: no matter what I do the default option is the lowest value.
		for (int i = 10; i < 310; i+=10)
		{
			char str[16];

			sprintf (str, "%d", i);
			gamma->addOption(str);
		}
		configmenu->appendChild(gamma);
		gamma->setValue("100");


		//      CheckboxMenu *fastsprites = new CheckboxMenu ("Fast sprites");
		//      fastsprites->setName("fastsprites");
		//      configmenu->appendChild(fastsprites);

		//
		// Configuration menu: Audio
		//

		SeparatorMenu *sepmenu = new SeparatorMenu("");
		configmenu->appendChild(sepmenu);

		CheckboxMenu *apuenable = new CheckboxMenu("Audio processor enable");
		apuenable->setName("apuenable");
		configmenu->appendChild(apuenable);

		CheckboxMenu *apumute = new CheckboxMenu("Audio processor mute");
		apumute->setName("apumute");
		configmenu->appendChild(apumute);

		SelectMenu *audio_quality = new SelectMenu ("Audio quality");
		audio_quality->setName("audio_quality");
		audio_quality->addOption("8000HzMono");
		audio_quality->addOption("11025HzMono");
		audio_quality->addOption("16000HzMono");
		audio_quality->addOption("22050HzMono");
		audio_quality->addOption("44100HzMono");
		audio_quality->addOption("8000HzStereo");
		audio_quality->addOption("11025HzStereo");
		audio_quality->addOption("16000HzStereo");
		audio_quality->addOption("22050HzStereo");
		audio_quality->addOption("44100HzStereo");
		configmenu->appendChild(audio_quality);
		audio_quality->setValue("16000HzStereo");

		CheckboxMenu *spc700hack = new CheckboxMenu("Audio performance hack");
		spc700hack->setName("spc700hack");
		configmenu->appendChild(spc700hack);

		//
		// Configuration menu: Miscellaneous
		//

		sepmenu = new SeparatorMenu("");
		configmenu->appendChild(sepmenu);

		CheckboxMenu *absromdir = new CheckboxMenu("Use absolute rom dir");
		absromdir->setName("absromdir");
		absromdir->setActionHandler(actionHandler);
		configmenu->appendChild(absromdir);

		CheckboxMenu *snesjoytype = new CheckboxMenu("SNES action buttons");
		snesjoytype->setName("snesjoytype");
		configmenu->appendChild(snesjoytype);

		SelectMenu *lowbattval = new SelectMenu ("Battery warning");
		lowbattval->setName("lowbattval");
		lowbattval->addOption("disabled");
		lowbattval->addOption("2.7V");
		lowbattval->addOption("2.6V");
		lowbattval->addOption("2.5V");
		lowbattval->addOption("2.4V");
		lowbattval->addOption("2.3V");
		lowbattval->addOption("2.2V");
		lowbattval->addOption("2.1V");
		lowbattval->addOption("2.0V");
		configmenu->appendChild(lowbattval);

		sepmenu = new SeparatorMenu("");
		configmenu->appendChild(sepmenu);

		// Save options //
		ActionMenu *saveopts = new ActionMenu("Save default config");
		saveopts->setActionHandler(actionHandler);
		saveopts->setConfirm(true, "Save current config?");
		saveopts->setName("saveoptions");
		configmenu->appendChild(saveopts);

		if (gameCfgName != NULL)
		{
			ActionMenu *saveopts_g = new ActionMenu("Save cfg for loaded ROM");
			saveopts_g->setActionHandler(actionHandler);
			saveopts_g->setConfirm(true, "Save config for current game?");
			saveopts_g->setName("saveoptions_game");
			configmenu->appendChild(saveopts_g);
		}

		// end of config menu

		ActionMenu *exitaction = new ActionMenu("Exit");
		exitaction->setIcon(new_image("skin/icons/exit.png"));
		exitaction->setActionHandler(actionHandler);
		exitaction->setConfirm(true,"");
		exitaction->setName("exitaction");
		menu->appendChild(exitaction);

	} catch (string message)
	{
		// probably an image that couldn't be found
		cout << "error building menu tree: " << message << endl;
		exit(2);
	}

	configFile = new ConfigFile();
	applyDefaultConfig(configFile);
	configFile->load("ssnes.cfg");
	menu->getValues(configFile);
	menu->setSelected (sel);
	menu->enter();
	parseconfig();

	// de-init system
	msystem.deInitSystem();

	// done
	delete configFile;
	delete actionHandler;
	delete surface;
	delete menu;
	// delete configmenu; // done by delete menu;
	delete_images();

	return &mr;
}

struct menureturn *skipmenu(const char *cfgfile)
{
	msystem.initSystem();
	configFile = new ConfigFile();
	configFile->load(cfgfile);
	parseconfig();
	msystem.deInitSystem();
	delete configFile;
	return &mr;
}

