#include "system.h"

#define i_hate_posix
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string>
#include <algorithm>

void checkbatt (void);

extern "C"
{
	#include "../gp2x_minimal.h"
};

using namespace std;

namespace hm 
{

   MenuSystem msystem;

	Event MenuSystem::pollEvent() 
	{
		// internal hm::Event
		Event event;
      unsigned long ke;
		static unsigned long oldke = 0;
		static EventKey oldek = KEY_NONE;
		static int nextevent = 0;

      event.type = EVENT_NONE;
      event.key = KEY_NONE;
      ke = gp2x_joystick_read();
      checkbatt();

		if (ke & GP2X_UP)
			event.key = KEY_UP;
		else if (ke & GP2X_LEFT)
			event.key = KEY_LEFT;
		else if (ke & GP2X_DOWN)
			event.key = KEY_DOWN;
		else if (ke & GP2X_RIGHT)
			event.key = KEY_RIGHT;
		else if (ke & GP2X_R)
			event.key = KEY_PGDOWN;
		else if (ke & GP2X_L)
			event.key = KEY_PGUP;
		else if (ke & GP2X_A)
			event.key = KEY_REJECT;
		else if (ke & GP2X_B)
			event.key = KEY_ACCEPT;
		else if (ke & GP2X_PUSH)
			event.key = KEY_ACCEPT;
		else
			ke = 0;			

		if ((ke != oldke) && (oldke != 0))
		{
			event.type = EVENT_KEYUP;
			event.key = oldek;
		}
		else if ((ke != oldke) && (oldke == 0))
		{
			event.type = EVENT_KEYDOWN;
		}
		else
		{
			// No key up or keydown, how about a timer event? Fire it every 30ms
			
			if (gp2x_timer_read() > nextevent)
			{
				nextevent = gp2x_timer_read() + 30;
				event.type = EVENT_TIMER;
			}
		}
		
		oldke = ke;
		oldek = event.key;

		return(event);

	}

   static bool mysort(FileDescription left, FileDescription right)
   {
      string l, r;

      l = left.name;
      r = right.name;

      for (int i = 0; i < l.length(); i ++)
         if (isupper(l[i]))
            l[i] = tolower (l[i]);

      for (int i = 0; i < r.length(); i ++)
         if (isupper(r[i]))
            r[i] = tolower (r[i]);

      return l < r;
   }

#ifdef i_hate_posix

	FileList MenuSystem::enumerateDirectory(string dirname) 
	{

		FileList list;
        DIR *handle;
        struct dirent *entry;
        struct stat info;

		// open dir
		handle = opendir(dirname.c_str());

		// if failed return an empty list
		if (handle == NULL) return(list);

        while ((entry = readdir(handle)) != NULL) 
        {
			FileDescription current;
			current.name = entry->d_name;
			string fullpath = dirname+"/"+entry->d_name;
            stat( fullpath.c_str() , &info );
			current.isdir = (info.st_mode & S_IFMT) == S_IFDIR;
			current.size = info.st_size;
			list.push_back(current);
        }

        // close directory
        closedir(handle);

		sort (list.begin(), list.end(), mysort);

		// sort, directories first
		FileList newList;
		for (unsigned int i=0; i<list.size(); i++)
			if (list.at(i).isdir) newList.push_back(list.at(i));
		for (unsigned int i=0; i<list.size(); i++)
			if (!list.at(i).isdir) newList.push_back(list.at(i));

		return(newList);

	}

#else

	FileList MenuSystem::enumerateDirectory(string dirname) 
	{

		FileList list;
		HANDLE handle;
		FileDescription current;
		WIN32_FIND_DATA entry;

		// if the thing we're accessing is not a directory, return empty list
		DWORD attrs = GetFileAttributes(dirname.c_str());
		if ( (attrs == 0xFFFFFFFF) || ((attrs && FILE_ATTRIBUTE_DIRECTORY) == 0))
			return(list);

		string full = dirname;

		// Circumvent a problem in FindFirstFile with c:\\* as parameter
		if ( (full.length() > 0) && (full[full.length()-1] != '\\') )
			full += "\\";

		// find first file
		handle = FindFirstFile( (full+"*").c_str(), &entry);

		if (handle == INVALID_HANDLE_VALUE)	
		{
			// none? return empty list
			return(list); 
		} 
		else 
		{
			while (true) 
			{
				// add file to the list
				current.name = entry.cFileName;
				current.isdir = entry.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY;
				current.size = entry.nFileSizeLow | (entry.nFileSizeHigh << 32);
				list.push_back(current);
				if (!FindNextFile(handle, &entry)) break;
			} 
		}
		
		sort (list.begin(), list.end(), mysort);

		// sort, directories first
		FileList newList;
		for (unsigned int i=0; i<list.size(); i++)
			if (list.at(i).isdir) newList.push_back(list.at(i));
		for (unsigned int i=0; i<list.size(); i++)
			if (!list.at(i).isdir) newList.push_back(list.at(i));

		return(newList);

	}

#endif 

	void MenuSystem::initSystem()
   {
	}

	void MenuSystem::deInitSystem() 
	{
	}
}

