#ifndef __system_h__
#define __system_h__

#include <string>
#include <vector>

using namespace std;

namespace hm {

	/**
		Event types
	*/
	typedef enum {
		EVENT_KEYDOWN,
		EVENT_KEYUP,
		EVENT_QUIT,
		EVENT_NONE,
		EVENT_TIMER
	} EventType;

	typedef enum  {
		KEY_UP,
		KEY_DOWN,
		KEY_LEFT,
		KEY_RIGHT,
		KEY_PGUP,
		KEY_PGDOWN,
		KEY_ACCEPT,
		KEY_REJECT,
		KEY_NONE
	} EventKey;

	/**
		Event object
	*/
	typedef struct {
		/**	Event type */
		EventType type;
		/** in case of key event, the key involved */
		EventKey key;
	} Event;

	/**
		File description object. Describes a single entry in a list of files. Used by enumerateDirectory
	*/
	typedef struct {
		/**	File name */
		string name;
		/**	Wether or not the file is a directory */
		bool isdir;
		/**	size */
		unsigned long long size;
	} FileDescription;

   typedef vector<FileDescription> FileList;

   class MenuSystem
   {
      public:

      /**	Checks if there's an event worthy of our attention and returns it. */
      Event pollEvent();

      /** returns a list of FileDescription objects representing the files in the given directory */
      FileList enumerateDirectory(string dirname);

      /** main screen turn on - should be called before doing anything */
      void initSystem();

      /** main screen turn off - should be called when you're done */
      void deInitSystem();
   };

   extern MenuSystem msystem;
};

#endif

